From d9c6caad5820f1943fe759476b04ba07f790b9bc Mon Sep 17 00:00:00 2001
From: Boris Smilga <boris.smilga@gmail.com>
Date: Mon, 16 Jul 2012 17:08:15 +0400
Subject: [PATCH 4/4] Check dynamically whether a RETURN operator has been invoked inside a loop.

This fixes incorrect handling of conditional returns, such as in

 (block nil
   (dotimes (i 10)
     (if (test i) (return i)))
   (return -1))
---
 src/compiler.lisp          |    6 ++++--
 src/non-cl.lisp            |    9 +++++----
 src/special-operators.lisp |    6 ++++--
 3 files changed, 13 insertions(+), 8 deletions(-)

diff --git a/src/compiler.lisp b/src/compiler.lisp
index 07a1eb7..0cb367c 100644
--- a/src/compiler.lisp
+++ b/src/compiler.lisp
@@ -85,8 +85,10 @@ block.")
   variables (otherwise they all share the same binding).")
 (defvar *loop-return-var* nil
   "Variable which is used to return values from inside loop bodies.")
-(defvar loop-returns? nil
-  "Set to T by RETURN-FROM when it returns a value from inside a loop.")
+(defvar *loop-return-set-var* nil
+  "Variable which is set by RETURN-FROM when it returns a value from inside
+  a loop.  The value is the name of a PS variable which dynamically
+  indicates if the return statement indeed has been invoked.")
 
 (defvar *loop-scope-lexicals*)
 (setf (documentation '*loop-scope-lexicals* 'variable)
diff --git a/src/non-cl.lisp b/src/non-cl.lisp
index 387bd29..ecfd96d 100644
--- a/src/non-cl.lisp
+++ b/src/non-cl.lisp
@@ -119,14 +119,15 @@
 (defmacro wrap-loop-for-return (name loop)
   (let ((loop-body (gensym)))
    `(let* ((*loop-return-var* nil)
-           (loop-returns? nil)
+           (*loop-return-set-var* nil)
            (,loop-body ,loop))
-      (if (and loop-returns? (not loop-wrapped?))
+      (if (and *loop-return-set-var* (not loop-wrapped?))
           (let ((loop-wrapped? t))
-            (compile-statement `(let (,*loop-return-var*)
+            (compile-statement `(let (,*loop-return-set-var* ,*loop-return-var*)
                                   (,',name ,@whole)
                                   ;; fix the below so it doesn't give a warning
-                                  (return ,*loop-return-var*))))
+                                  (when ,*loop-return-set-var*
+                                    (return ,*loop-return-var*)))))
           ,loop-body))))
 
 (define-statement-operator for (init-forms cond-forms step-forms &body body)
diff --git a/src/special-operators.lisp b/src/special-operators.lisp
index febfd18..8ac4025 100644
--- a/src/special-operators.lisp
+++ b/src/special-operators.lisp
@@ -283,10 +283,12 @@ Parenscript now implements implicit return, update your code! Things like (lambd
                  ((eq 'values (car form)) (return-exp tag (cadr form) (cddr form)))
                  (t                       (expressionize-result tag form)))))
         (in-loop-scope?
-         (setf loop-returns?     t
+         (setf *loop-return-set-var* (or *loop-return-set-var*
+                                         (ps-gensym "loop-result-var-set"))
                *loop-return-var* (or *loop-return-var*
                                      (ps-gensym "loop-result-var")))
-         (compile-statement `(progn (setf ,*loop-return-var* ,result)
+         (compile-statement `(progn (setf ,*loop-return-set-var* t
+                                          ,*loop-return-var* ,result)
                                     (break))))
         (t
          (ps-compile `(return-from nilBlock ,result)))))
-- 
1.7.3.2

